; MySQL's Connector/C ships with a libmysql.dll main library and libmysql.lib
; interface library. However, the interface library is not linkable by MinGW.
;
; At compile time, we generate a libmysql.a interface library with dlltool.exe.
;
; This def file can be re-generated using the reimp.exe or gendef.exe tools.
;
LIBRARY libmysql.dll
EXPORTS
mysql_affected_rows
mysql_affected_rows@4
mysql_change_user
mysql_change_user@16
mysql_character_set_name
mysql_character_set_name@4
mysql_close
mysql_close@4
mysql_data_seek
mysql_data_seek@12
mysql_debug
mysql_debug@4
mysql_dump_debug_info
mysql_dump_debug_info@4
mysql_eof
mysql_eof@4
mysql_errno
mysql_errno@4
mysql_error
mysql_error@4
mysql_escape_string
mysql_escape_string@12
mysql_fetch_field
mysql_fetch_field@4
mysql_fetch_field_direct
mysql_fetch_field_direct@8
mysql_fetch_fields
mysql_fetch_fields@4
mysql_fetch_lengths
mysql_fetch_lengths@4
mysql_fetch_row
mysql_fetch_row@4
mysql_field_count
mysql_field_count@4
mysql_field_seek
mysql_field_seek@8
mysql_field_tell
mysql_field_tell@4
mysql_free_result
mysql_free_result@4
mysql_get_client_info
mysql_get_client_info@0
mysql_get_client_version
mysql_get_client_version@0
mysql_get_host_info
mysql_get_host_info@4
mysql_get_option
mysql_get_option@12
mysql_get_proto_info
mysql_get_proto_info@4
mysql_get_server_info
mysql_get_server_info@4
mysql_get_server_version
mysql_get_server_version@4
mysql_get_ssl_cipher
mysql_get_ssl_cipher@4
mysql_hex_string
mysql_hex_string@12
mysql_info
mysql_info@4
mysql_init
mysql_init@4
mysql_insert_id
mysql_insert_id@4
mysql_kill
mysql_kill@8
mysql_library_end
mysql_library_end@0
mysql_library_init
mysql_library_init@12
mysql_list_dbs
mysql_list_dbs@8
mysql_list_fields
mysql_list_fields@12
mysql_list_processes
mysql_list_processes@4
mysql_list_tables
mysql_list_tables@8
mysql_more_results
mysql_more_results@4
mysql_next_result
mysql_next_result@4
mysql_num_fields
mysql_num_fields@4
mysql_num_rows
mysql_num_rows@4
mysql_options
mysql_options@12
mysql_options4
mysql_options4@16
mysql_ping
mysql_ping@4
mysql_query
mysql_query@8
mysql_read_query_result
mysql_read_query_result@4
mysql_real_connect
mysql_real_connect@32
mysql_real_escape_string
mysql_real_escape_string@16
mysql_real_query
mysql_real_query@12
mysql_refresh
mysql_refresh@8
mysql_reset_connection
mysql_reset_connection@4
mysql_rollback
mysql_rollback@4
mysql_row_seek
mysql_row_seek@8
mysql_row_tell
mysql_row_tell@4
mysql_select_db
mysql_select_db@8
mysql_send_query
mysql_send_query@12
mysql_server_end
mysql_server_end@0
mysql_server_init
mysql_server_init@12
mysql_session_track_get_first
mysql_session_track_get_first@16
mysql_session_track_get_next
mysql_session_track_get_next@16
mysql_set_character_set
mysql_set_character_set@8
mysql_set_local_infile_default
mysql_set_local_infile_default@4
mysql_set_local_infile_handler
mysql_set_local_infile_handler@24
mysql_set_server_option
mysql_set_server_option@8
mysql_shutdown
mysql_shutdown@8
mysql_sqlstate
mysql_sqlstate@4
mysql_ssl_set
mysql_ssl_set@24
mysql_stat
mysql_stat@4
mysql_stmt_affected_rows
mysql_stmt_affected_rows@4
mysql_stmt_attr_get
mysql_stmt_attr_get@12
mysql_stmt_attr_set
mysql_stmt_attr_set@12
mysql_stmt_bind_param
mysql_stmt_bind_param@8
mysql_stmt_bind_result
mysql_stmt_bind_result@8
mysql_stmt_close
mysql_stmt_close@4
mysql_stmt_data_seek
mysql_stmt_data_seek@12
mysql_stmt_errno
mysql_stmt_errno@4
mysql_stmt_error
mysql_stmt_error@4
mysql_stmt_execute
mysql_stmt_execute@4
mysql_stmt_fetch
mysql_stmt_fetch@4
mysql_stmt_fetch_column
mysql_stmt_fetch_column@16
mysql_stmt_field_count
mysql_stmt_field_count@4
mysql_stmt_free_result
mysql_stmt_free_result@4
mysql_stmt_init
mysql_stmt_init@4
mysql_stmt_insert_id
mysql_stmt_insert_id@4
mysql_stmt_next_result
mysql_stmt_next_result@4
mysql_stmt_num_rows
mysql_stmt_num_rows@4
mysql_stmt_param_count
mysql_stmt_param_count@4
mysql_stmt_param_metadata
mysql_stmt_param_metadata@4
mysql_stmt_prepare
mysql_stmt_prepare@12
mysql_stmt_reset
mysql_stmt_reset@4
mysql_stmt_result_metadata
mysql_stmt_result_metadata@4
mysql_stmt_row_seek
mysql_stmt_row_seek@8
mysql_stmt_row_tell
mysql_stmt_row_tell@4
mysql_stmt_send_long_data
mysql_stmt_send_long_data@16
mysql_stmt_sqlstate
mysql_stmt_sqlstate@4
mysql_stmt_store_result
mysql_stmt_store_result@4
mysql_store_result
mysql_store_result@4
mysql_thread_end
mysql_thread_end@0
mysql_thread_id
mysql_thread_id@4
mysql_thread_init
mysql_thread_init@0
mysql_thread_safe
mysql_thread_safe@0
mysql_use_result
mysql_use_result@4
mysql_warning_count
mysql_warning_count@4
